from urllib.error import URLError
from urllib.parse import urlencode

from django.http import HttpResponseRedirect

from .. import PaymentError
from .. import PaymentStatus
from .. import RedirectNeeded
from ..core import BasicProvider
from .forms import DummyForm


class DummyProvider(BasicProvider):
    """Dummy payment provider.

    This is a dummy backend suitable for testing your store without contacting any
    payment gateways. Instead of using an external service it will simply show you a
    form that allows you to confirm or reject the payment.
    """

    def get_form(self, payment, data=None):
        if payment.status == PaymentStatus.WAITING:
            payment.change_status(PaymentStatus.INPUT)
        form = DummyForm(data=data, hidden_inputs=False, provider=self,
                         payment=payment)
        if form.is_valid():
            new_status = form.cleaned_data['status']
            payment.change_status(new_status)
            new_fraud_status = form.cleaned_data['fraud_status']
            payment.change_fraud_status(new_fraud_status)

            gateway_response = form.cleaned_data.get('gateway_response')
            verification_result = form.cleaned_data.get('verification_result')
            if gateway_response or verification_result:
                if gateway_response == '3ds-disabled':
                    # Standard request without 3DSecure
                    pass
                elif gateway_response == '3ds-redirect':
                    # Simulate redirect to 3DS and get back to normal
                    # payment processing
                    process_url = payment.get_process_url()
                    params = urlencode(
                        {'verification_result': verification_result})
                    redirect_url = f'{process_url}?{params}'
                    raise RedirectNeeded(redirect_url)
                elif gateway_response == 'failure':
                    # Gateway raises error (HTTP 500 for example)
                    raise URLError('Opps')
                elif gateway_response == 'payment-error':
                    raise PaymentError('Unsupported operation')

            if new_status in [PaymentStatus.PREAUTH, PaymentStatus.CONFIRMED]:
                raise RedirectNeeded(payment.get_success_url())
            raise RedirectNeeded(payment.get_failure_url())
        return form

    def process_data(self, payment, request):
        verification_result = request.GET.get('verification_result')
        if verification_result:
            payment.change_status(verification_result)
        if payment.status in [PaymentStatus.CONFIRMED, PaymentStatus.PREAUTH]:
            return HttpResponseRedirect(payment.get_success_url())
        return HttpResponseRedirect(payment.get_failure_url())

    def capture(self, payment, amount=None):
        payment.change_status(PaymentStatus.CONFIRMED)
        return amount

    def release(self, payment):
        return None

    def refund(self, payment, amount=None):
        return amount or 0



import hashlib

from django import forms

from .. import PaymentStatus


NEW = 'new'
PROCESSING = 'processing'
COMPLETED = 'completed'
REJECTED = 'rejected'
PROCESSING_REALIZATION_WAITING = 'processing_realization_waiting'
PROCESSING_REALIZATION = 'processing_realization'


class ProcessPaymentForm(forms.Form):

    id = forms.CharField(required=False)
    operation_number = forms.CharField(required=False)
    operation_type = forms.CharField(required=False)
    operation_status = forms.CharField(required=False)
    operation_amount = forms.CharField(required=False)
    operation_currency = forms.CharField(required=False)
    operation_withdrawal_amount = forms.CharField(required=False)
    operation_commission_amount = forms.CharField(required=False)
    is_completed = forms.CharField(required=False)
    operation_original_amount = forms.CharField(required=False)
    operation_original_currency = forms.CharField(required=False)
    operation_datetime = forms.CharField(required=False)
    operation_related_number = forms.CharField(required=False)
    control = forms.CharField(required=False)
    description = forms.CharField(required=False)
    email = forms.CharField(required=False)
    p_info = forms.CharField(required=False)
    p_email = forms.CharField(required=False)
    credit_card_issuer_identification_number = forms.CharField(required=False)
    credit_card_masked_number = forms.CharField(required=False)
    credit_card_brand_codename = forms.CharField(required=False)
    credit_card_brand_code = forms.CharField(required=False)
    credit_card_id = forms.CharField(required=False)
    channel = forms.CharField(required=False)
    channel_country = forms.CharField(required=False)
    geoip_country = forms.CharField(required=False)
    signature = forms.CharField(required=True)

    def __init__(self, payment, pin, **kwargs):
        super().__init__(**kwargs)
        self.pin = pin
        self.payment = payment

    def clean(self):
        cleaned_data = super().clean()
        if not self.errors:
            key_vars = (
                self.pin,
                cleaned_data.get('id', ''),
                cleaned_data.get('operation_number', ''),
                cleaned_data.get('operation_type', ''),
                cleaned_data.get('operation_status', ''),
                cleaned_data.get('operation_amount', ''),
                cleaned_data.get('operation_currency', ''),
                cleaned_data.get('operation_withdrawal_amount', ''),
                cleaned_data.get('operation_commission_amount', ''),
                cleaned_data.get('is_completed', ''),
                cleaned_data.get('operation_original_amount', ''),
                cleaned_data.get('operation_original_currency', ''),
                cleaned_data.get('operation_datetime', ''),
                cleaned_data.get('operation_related_number', ''),
                cleaned_data.get('control', ''),
                cleaned_data.get('description', ''),
                cleaned_data.get('email', ''),
                cleaned_data.get('p_info', ''),
                cleaned_data.get('p_email', ''),
                cleaned_data.get('credit_card_issuer_identification_number', ''),
                cleaned_data.get('credit_card_masked_number', ''),
                cleaned_data.get('credit_card_brand_codename', ''),
                cleaned_data.get('credit_card_brand_code', ''),
                cleaned_data.get('credit_card_id', ''),
                cleaned_data.get('channel', ''),
                cleaned_data.get('channel_country', ''),
                cleaned_data.get('geoip_country', ''))
            key = ''.join(key_vars)
            sha256 = hashlib.sha256()
            sha256.update(key.encode('utf-8'))
            key_hash = sha256.hexdigest()
            if key_hash != self.cleaned_data['signature']:
                self._errors['signature'] = self.error_class(['Bad hash'])
            if int(cleaned_data['control']) != self.payment.id:
                self._errors['control'] = self.error_class(['Bad payment id'])
        return cleaned_data

    def save(self, *args, **kwargs):
        status = self.cleaned_data['operation_status']
        self.payment.transaction_id = self.cleaned_data['operation_number']
        self.payment.save()
        if status == COMPLETED:
            self.payment.captured_amount = self.payment.total
            self.payment.change_status(PaymentStatus.CONFIRMED)
        elif status == REJECTED:
            self.payment.change_status(PaymentStatus.REJECTED)



import json
from decimal import Decimal
from unittest import TestCase
from unittest.mock import MagicMock
from unittest.mock import Mock
from unittest.mock import patch

from django.utils import timezone
from requests import HTTPError

from . import PaypalCardProvider
from . import PaypalProvider
from .. import PaymentError
from .. import PaymentStatus
from .. import PurchasedItem
from .. import RedirectNeeded

CLIENT_ID = 'abc123'
PAYMENT_TOKEN = '5a4dae68-2715-4b1e-8bb2-2c2dbe9255f6'
SECRET = '123abc'
VARIANT = 'wallet'

PROCESS_DATA = {
    'name': 'John Doe',
    'number': '371449635398431',
    'expiration_0': '5',
    'expiration_1': '2023',
    'cvv2': '1234'}


class Payment(Mock):
    id = 1
    description = 'payment'
    currency = 'USD'
    delivery = Decimal(10)
    status = PaymentStatus.WAITING
    tax = Decimal(10)
    token = PAYMENT_TOKEN
    total = Decimal(220)
    captured_amount = Decimal(0)
    variant = VARIANT
    transaction_id = None
    message = ''
    extra_data = json.dumps({'links': {
        'approval_url': None,
        'capture': {'href': 'http://capture.com'},
        'refund': {'href': 'http://refund.com'},
        'execute': {'href': 'http://execute.com'}
    }})

    def change_status(self, status, message=''):
        self.status = status
        self.message = message

    def get_failure_url(self):
        return 'http://cancel.com'

    def get_process_url(self):
        return 'http://example.com'

    def get_purchased_items(self):
        return [
            PurchasedItem(
                name='foo', quantity=10, price=Decimal('20'),
                currency='USD', sku='bar')]

    def get_success_url(self):
        return 'http://success.com'


class TestPaypalProvider(TestCase):

    def setUp(self):
        self.payment = Payment()
        self.provider = PaypalProvider(secret=SECRET, client_id=CLIENT_ID)

    def test_provider_raises_redirect_needed_on_success(self):
        with patch('requests.post') as mocked_post:
            transaction_id = '1234'
            data = MagicMock()
            data.return_value = {
                'id': transaction_id,
                'token_type': 'test_token_type',
                'access_token': 'test_access_token',
                'links': [
                    {'rel': 'approval_url', 'href': 'http://approval_url.com'}
                ]}
            post = MagicMock()
            post.json = data
            post.status_code = 200
            mocked_post.return_value = post
            with self.assertRaises(RedirectNeeded):
                self.provider.get_form(payment=self.payment)
        self.assertEqual(self.payment.status, PaymentStatus.WAITING)
        self.assertEqual(self.payment.captured_amount, Decimal('0'))
        self.assertEqual(self.payment.transaction_id, transaction_id)

    @patch('requests.post')
    def test_provider_captures_payment(self, mocked_post):
        data = MagicMock()
        data.return_value = {
            'state': 'completed',
            'token_type': 'test_token_type',
            'access_token': 'test_access_token'}
        post = MagicMock()
        post.json = data
        post.status_code = 200
        mocked_post.return_value = post
        self.provider.capture(self.payment)
        self.assertEqual(self.payment.status, PaymentStatus.CONFIRMED)

    @patch('requests.post')
    def test_provider_handles_captured_payment(self, mocked_post):
        data = MagicMock()
        data.return_value = {
            'name': 'AUTHORIZATION_ALREADY_COMPLETED'}
        response = MagicMock()
        response.json = data
        mocked_post.side_effect = HTTPError(response=response)
        self.provider.capture(self.payment)
        self.assertEqual(self.payment.status, PaymentStatus.CONFIRMED)

    @patch('requests.post')
    def test_provider_refunds_payment(self, mocked_post):
        data = MagicMock()
        data.return_value = {
            'token_type': 'test_token_type',
            'access_token': 'test_access_token'}
        post = MagicMock()
        post.json = data
        post.status_code = 200
        mocked_post.return_value = post
        self.provider.refund(self.payment)
        self.assertEqual(self.payment.status, PaymentStatus.REFUNDED)

    @patch('requests.post')
    @patch('payments.paypal.redirect')
    def test_provider_redirects_on_success_captured_payment(
            self, mocked_redirect, mocked_post):
        data = MagicMock()
        data.return_value = {
            'token_type': 'test_token_type',
            'access_token': 'test_access_token',
            'payer': {'payer_info': 'test123'},
            'transactions': [
                {'related_resources': [{
                    'sale': {'links': ''},
                    'authorization': {'links': ''}}]}
            ]}
        post = MagicMock()
        post.json = data
        post.status_code = 200
        mocked_post.return_value = post

        request = MagicMock()
        request.GET = {'token': 'test', 'PayerID': '1234'}
        self.provider.process_data(self.payment, request)

        self.assertEqual(self.payment.status, PaymentStatus.CONFIRMED)
        self.assertEqual(self.payment.captured_amount, self.payment.total)

    @patch('requests.post')
    @patch('payments.paypal.redirect')
    def test_provider_redirects_on_success_preauth_payment(
            self, mocked_redirect, mocked_post):
        data = MagicMock()
        data.return_value = {
            'token_type': 'test_token_type',
            'access_token': 'test_access_token',
            'payer': {'payer_info': 'test123'},
            'transactions': [
                {'related_resources': [{
                    'sale': {'links': ''},
                    'authorization': {'links': ''}}]}
            ]}
        post = MagicMock()
        post.json = data
        post.status_code = 200
        mocked_post.return_value = post

        request = MagicMock()
        request.GET = {'token': 'test', 'PayerID': '1234'}
        provider = PaypalProvider(
            secret=SECRET, client_id=CLIENT_ID, capture=False)
        provider.process_data(self.payment, request)

        self.assertEqual(self.payment.status, PaymentStatus.PREAUTH)
        self.assertEqual(self.payment.captured_amount, Decimal('0'))

    @patch('payments.paypal.redirect')
    def test_provider_request_without_payerid_redirects_on_failure(
            self, mocked_redirect):
        request = MagicMock()
        request.GET = {'token': 'test', 'PayerID': None}
        self.provider.process_data(self.payment, request)
        self.assertEqual(self.payment.status, PaymentStatus.REJECTED)

    @patch('requests.post')
    def test_provider_renews_access_token(self, mocked_post):
        new_token = 'new_test_token'
        response401 = MagicMock()
        response401.status_code = 401
        data = MagicMock()
        data.return_value = {'access_token': new_token, 'token_type': 'type'}
        response = MagicMock()
        response.json = data
        response.status_code = 200
        mocked_post.side_effect = [
            HTTPError(response=response401), response, response]

        self.payment.created = timezone.now()
        self.payment.extra_data = json.dumps({
            'auth_response': {
                'access_token': 'expired_token',
                'token_type': 'token type',
                'expires_in': 99999}})
        self.provider.create_payment(self.payment)
        payment_response = json.loads(self.payment.extra_data)['auth_response']
        self.assertEqual(payment_response['access_token'], new_token)


class TestPaypalCardProvider(TestCase):

    def setUp(self):
        self.payment = Payment(extra_data='')
        self.provider = PaypalCardProvider(secret=SECRET, client_id=CLIENT_ID)

    def test_provider_raises_redirect_needed_on_success_captured_payment(self):
        with patch('requests.post') as mocked_post:
            transaction_id = '1234'
            data = MagicMock()
            data.return_value = {
                'id': transaction_id,
                'token_type': 'test_token_type',
                'access_token': 'test_access_token',
                'transactions': [
                    {'related_resources': [
                        {'sale': {'links': [
                            {'rel': 'refund', 'href': 'http://refund.com'}]}}
                    ]}
                ]}
            post = MagicMock()
            post.json = data
            post.status_code = 200
            mocked_post.return_value = post
            with self.assertRaises(RedirectNeeded) as exc:
                self.provider.get_form(
                    payment=self.payment, data=PROCESS_DATA)
                self.assertEqual(exc.args[0], self.payment.get_success_url())
        links = self.provider._get_links(self.payment)
        self.assertEqual(self.payment.status, PaymentStatus.CONFIRMED)
        self.assertEqual(self.payment.captured_amount, self.payment.total)
        self.assertEqual(self.payment.transaction_id, transaction_id)
        self.assertTrue('refund' in links)

    def test_provider_raises_redirect_needed_on_success_preauth_payment(self):
        provider = PaypalCardProvider(
            secret=SECRET, client_id=CLIENT_ID, capture=False)
        with patch('requests.post') as mocked_post:
            transaction_id = '1234'
            data = MagicMock()
            data.return_value = {
                'id': transaction_id,
                'token_type': 'test_token_type',
                'access_token': 'test_access_token',
                'transactions': [
                    {'related_resources': [
                        {'authorization': {'links': [
                            {'rel': 'refund', 'href': 'http://refund.com'},
                            {'rel': 'capture', 'href': 'http://capture.com'}]}}
                    ]}
                ]}
            post = MagicMock()
            post.json = data
            post.status_code = 200
            mocked_post.return_value = post
            with self.assertRaises(RedirectNeeded) as exc:
                provider.get_form(
                    payment=self.payment, data=PROCESS_DATA)
                self.assertEqual(exc.args[0], self.payment.get_success_url())
        links = provider._get_links(self.payment)
        self.assertEqual(self.payment.status, PaymentStatus.PREAUTH)
        self.assertEqual(self.payment.captured_amount, Decimal('0'))
        self.assertEqual(self.payment.transaction_id, transaction_id)
        self.assertTrue('capture' in links)
        self.assertTrue('refund' in links)

    def test_form_shows_validation_error_message(self):
        with patch('requests.post') as mocked_post:
            error_message = 'error message'
            data = MagicMock()
            data.return_value = {'details': [{'issue': error_message}]}
            post = MagicMock()
            post.json = data
            post.status_code = 400
            mocked_post.side_effect = HTTPError(response=post)
            form = self.provider.get_form(
                payment=self.payment, data=PROCESS_DATA)
        self.assertEqual(self.payment.status, PaymentStatus.ERROR)
        self.assertEqual(form.errors['__all__'][0], error_message)

    def test_form_shows_internal_error_message(self):
        with patch('requests.post') as mocked_post:
            error_message = 'error message'
            data = MagicMock()
            data.return_value = {
                'token_type': 'test_token_type',
                'access_token': 'test_access_token',
                'message': error_message}
            post = MagicMock()
            post.status_code = 400
            post.json = data
            mocked_post.return_value = post
            with self.assertRaises(PaymentError):
                self.provider.get_form(
                    payment=self.payment, data=PROCESS_DATA)
        self.assertEqual(self.payment.status, PaymentStatus.ERROR)
        self.assertEqual(self.payment.message, error_message)



import datetime
import os.path

import suds.client
import suds.wsse
from django.core import signing
from django.shortcuts import redirect
from django.utils.translation import ugettext as _
from suds.sax.element import Element
from suds.sudsobject import Object

from .. import ExternalPostNeeded
from .. import FraudStatus
from .. import PaymentError
from .. import PaymentStatus
from .. import RedirectNeeded
from ..core import BasicProvider
from ..core import get_credit_card_issuer
from ..forms import PaymentForm as BaseForm
from .forms import PaymentForm


ACCEPTED = 100
TRANSACTION_SETTLED = 238
TRANSACTION_REVERSED = 237
AUTHENTICATE_REQUIRED = 475

FRAUD_MANAGER_REVIEW = 480
FRAUD_MANAGER_REJECT = 481
FRAUD_SCORE_EXCEEDS_THRESHOLD = 400

# Soft Decline
ADDRESS_VERIFICATION_SERVICE_FAIL = 200
CARD_VERIFICATION_NUMBER_FAIL = 230
SMART_AUTHORIZATION_FAIL = 520

WSDL_PATH_TEST = 'xml/CyberSourceTransaction_1.101.test.wsdl'
WSDL_PATH = 'xml/CyberSourceTransaction_1.101.wsdl'


class CyberSourceProvider(BasicProvider):
    """Payment provider for CyberSource

    This backend implements payments using `Cybersource
    <http://www.cybersource.com/www/>`_.

    This backend supports fraud detection.

    :param merchant_id: Your Merchant ID
    :param password: Generated transaction security key for the SOAP toolkit
    :param org_id: Provide this parameter to enable Cybersource Device Fingerprinting
    :param fingerprint_url: Address of the fingerprint server
    :param sandbox: Whether to use a sandbox environment for testing
    :param capture: Whether to capture the payment automatically.  See
        :ref:`capture-payments` for more details.
    """

    fingerprint_url: str

    def __init__(
        self, merchant_id, password, org_id=None,
        fingerprint_url='https://h.online-metrix.net/fp/', sandbox=True,
        capture=True,
    ):
        self.merchant_id = merchant_id
        self.password = password
        local_path = os.path.dirname(__file__)
        if os.path.sep != '/':
            # ugly hack for urllib and Windows
            local_path = local_path.replace(os.path.sep, '/')
        if not local_path.startswith('/'):
            # windows paths don't start with '/'
            local_path = f'/{local_path}'
        if sandbox:
            wsdl_path = f'file://{local_path}/{WSDL_PATH_TEST}'
            self.endpoint = (
                'https://ics2wstest.ic3.com/commerce/1.x/transactionProcessor')
        else:
            wsdl_path = f'file://{local_path}/{WSDL_PATH}'
            self.endpoint = (
                'https://ics2ws.ic3.com/commerce/1.x/transactionProcessor')
        self.client = suds.client.Client(wsdl_path)
        self.fingerprint_url = fingerprint_url
        self.org_id = org_id
        security_header = suds.wsse.Security()
        security_token = suds.wsse.UsernameToken(
            username=self.merchant_id,
            password=self.password)
        security_header.tokens.append(security_token)
        self.client.set_options(soapheaders=[security_header.xml()])
        super().__init__(capture=capture)

    def get_form(self, payment, data=None):
        if payment.status == PaymentStatus.WAITING:
            payment.change_status(PaymentStatus.INPUT)
        form = PaymentForm(data, provider=self, payment=payment)
        try:
            if form.is_valid():
                raise RedirectNeeded(payment.get_success_url())
        except ExternalPostNeeded as e:
            return e.args[0]
        return form

    def _change_status_to_confirmed(self, payment):
        if self._capture:
            payment.captured_amount = payment.total
            payment.change_status(PaymentStatus.CONFIRMED)
        else:
            payment.change_status(PaymentStatus.PREAUTH)

    def _set_proper_payment_status_from_reason_code(self, payment, reason_code):
        if reason_code == ACCEPTED:
            payment.change_fraud_status(FraudStatus.ACCEPT, commit=False)
            self._change_status_to_confirmed(payment)
        elif reason_code == FRAUD_MANAGER_REVIEW:
            payment.change_fraud_status(
                FraudStatus.REVIEW,
                _('The order is marked for review by Decision Manager'),
                commit=False)
            self._change_status_to_confirmed(payment)
        elif reason_code == FRAUD_MANAGER_REJECT:
            payment.change_fraud_status(
                FraudStatus.REJECT, _('The order has been rejected by Decision Manager'),
                commit=False)
            self._change_status_to_confirmed(payment)
        elif reason_code == FRAUD_SCORE_EXCEEDS_THRESHOLD:
            payment.change_fraud_status(
                FraudStatus.REJECT, _('Fraud score exceeds threshold.'), commit=False)
            self._change_status_to_confirmed(payment)
        elif reason_code == SMART_AUTHORIZATION_FAIL:
            payment.change_fraud_status(
                FraudStatus.REJECT, _('CyberSource Smart Authorization failed.'),
                commit=False)
            self._change_status_to_confirmed(payment)
        elif reason_code == CARD_VERIFICATION_NUMBER_FAIL:
            payment.change_fraud_status(
                FraudStatus.REJECT, _('Card verification number (CVN) did not match.'),
                commit=False)
            self._change_status_to_confirmed(payment)
        elif reason_code == ADDRESS_VERIFICATION_SERVICE_FAIL:
            payment.change_fraud_status(
                FraudStatus.REJECT, _(
                    'CyberSource Address Verification Service failed.'),
                commit=False)
            self._change_status_to_confirmed(payment)
        else:
            error = self._get_error_message(reason_code)
            payment.change_status(PaymentStatus.ERROR, message=error)
            raise PaymentError(error)

    def charge(self, payment, data):
        if self._capture:
            params = self._prepare_sale(payment, data)
        else:
            params = self._prepare_preauth(payment, data)
        response = self._make_request(payment, params)
        payment.attrs.capture = self._capture
        payment.transaction_id = response.requestID
        if response.reasonCode == AUTHENTICATE_REQUIRED:
            xid = response.payerAuthEnrollReply.xid
            payment.attrs.xid = xid
            payment.change_status(
                PaymentStatus.WAITING,
                message=_('3-D Secure verification in progress'))
            action = response.payerAuthEnrollReply.acsURL
            cc_data = dict(data)
            expiration = cc_data.pop('expiration')
            cc_data['expiration'] = {
                'month': expiration.month,
                'year': expiration.year}
            cc_data = signing.dumps(cc_data)
            payload = {
                'PaReq': response.payerAuthEnrollReply.paReq,
                'TermUrl': self.get_return_url(payment, {'token': cc_data}),
                'MD': xid}
            form = BaseForm(data=payload, action=action, autosubmit=True)
            raise ExternalPostNeeded(form)
        else:
            self._set_proper_payment_status_from_reason_code(
                payment, response.reasonCode)

    def capture(self, payment, amount=None):
        if amount is None:
            amount = payment.total
        params = self._prepare_capture(payment, amount=amount)
        response = self._make_request(payment, params)
        if response.reasonCode == ACCEPTED:
            payment.transaction_id = response.requestID
        elif response.reasonCode == TRANSACTION_SETTLED:
            payment.change_status(PaymentStatus.CONFIRMED)
        else:
            payment.save()
            error = self._get_error_message(response.reasonCode)
            raise PaymentError(error)
        return amount

    def release(self, payment):
        params = self._prepare_release(payment)
        response = self._make_request(payment, params)
        if response.reasonCode == ACCEPTED:
            payment.transaction_id = response.requestID
        elif response.reasonCode != TRANSACTION_REVERSED:
            payment.save()
            error = self._get_error_message(response.reasonCode)
            raise PaymentError(error)

    def refund(self, payment, amount=None):
        if amount is None:
            amount = payment.captured_amount
        params = self._prepare_refund(payment, amount=amount)
        response = self._make_request(payment, params)
        payment.save()
        if response.reasonCode != ACCEPTED:
            error = self._get_error_message(response.reasonCode)
            raise PaymentError(error)
        return amount

    def _get_error_message(self, code):
        if code in [221, 222, 700, 701, 702, 703]:
            return _(
                'Our bank has flagged your transaction as unusually suspicious. Please contact us to resolve this issue.')  # noqa
        elif code in [201, 203, 209]:
            return _(
                'Your bank has declined the transaction. No additional information was provided.')  # noqa
        elif code == 202:
            return _(
                'The card has either expired or you have entered an incorrect expiration date.')  # noqa
        elif code in [204, 210, 251]:
            return _(
                'There are insufficient funds on your card or it has reached its credit limit.')  # noqa
        elif code == 205:
            return _(
                'The card you are trying to use was reported as lost or stolen.')  # noqa
        elif code == 208:
            return _(
                'Your card is either inactive or it does not permit online payments. Please contact your bank to resolve this issue.')  # noqa
        elif code == 211:
            return _(
                'Your bank has declined the transaction. Please check the verification number of your card and retry.')  # noqa
        elif code == 231:
            return _(
                'Your bank has declined the transaction. Please make sure the card number you have entered is correct and retry.')  # noqa
        elif code in [232, 240]:
            return _(
                'We are sorry but our bank cannot handle the card type you are using.')  # noqa
        elif code in [450, 451, 452, 453, 454, 455,
                      456, 457, 458, 459, 460, 461]:
            return _(
                'We were unable to verify your address. Please make sure the address you entered is correct and retry.')  # noqa
        else:
            return _(
                'We were unable to complete the transaction. Please try again later.')  # noqa

    def _get_params_for_new_payment(self, payment):
        params = {
            'merchantID': self.merchant_id,
            'merchantReferenceCode': payment.id,
        }
        try:
            fingerprint_id = payment.attrs.fingerprint_session_id
        except KeyError:
            pass
        else:
            params['deviceFingerprintID'] = fingerprint_id
        merchant_defined_data = self._prepare_merchant_defined_data(payment)
        if merchant_defined_data:
            params['merchantDefinedData'] = merchant_defined_data
        return params

    def _make_request(self, payment, params):
        response = self.client.service.runTransaction(**params)
        payment.attrs.last_response = self._serialize_response(response)
        return response

    def _prepare_payer_auth_validation_check(self, payment, card_data,
                                             pa_response):
        check_service = self.client.factory.create(
            'data:PayerAuthValidateService')
        check_service._run = 'true'
        check_service.signedPARes = pa_response
        params = self._get_params_for_new_payment(payment)
        params['payerAuthValidateService'] = check_service
        if payment.attrs.capture:
            service = self.client.factory.create('data:CCCreditService')
            service._run = 'true'
            params['ccCreditService'] = service
        else:
            service = self.client.factory.create('data:CCAuthService')
            service._run = 'true'
            params['ccAuthService'] = service
        params.update({
            'billTo': self._prepare_billing_data(payment),
            'card': self._prepare_card_data(card_data),
            'item': self._prepare_items(payment),
            'purchaseTotals': self._prepare_totals(payment)})
        return params

    def _prepare_sale(self, payment, card_data):
        service = self.client.factory.create('data:CCCreditService')
        service._run = 'true'
        check_service = self.client.factory.create(
            'data:PayerAuthEnrollService')
        check_service._run = 'true'
        params = self._get_params_for_new_payment(payment)
        params.update({
            'ccCreditService': service,
            'payerAuthEnrollService': check_service,
            'billTo': self._prepare_billing_data(payment),
            'card': self._prepare_card_data(card_data),
            'item': self._prepare_items(payment),
            'purchaseTotals': self._prepare_totals(payment)})
        return params

    def _prepare_preauth(self, payment, card_data):
        service = self.client.factory.create('data:CCAuthService')
        service._run = 'true'
        check_service = self.client.factory.create(
            'data:PayerAuthEnrollService')
        check_service._run = 'true'
        params = self._get_params_for_new_payment(payment)
        params.update({
            'ccAuthService': service,
            'payerAuthEnrollService': check_service,
            'billTo': self._prepare_billing_data(payment),
            'card': self._prepare_card_data(card_data),
            'item': self._prepare_items(payment),
            'purchaseTotals': self._prepare_totals(payment)})
        return params

    def _prepare_capture(self, payment, amount=None):
        service = self.client.factory.create('data:CCCaptureService')
        service._run = 'true'
        service.authRequestID = payment.transaction_id
        params = {
            'merchantID': self.merchant_id,
            'merchantReferenceCode': payment.id,
            'ccCaptureService': service,
            'purchaseTotals': self._prepare_totals(payment, amount=amount)}
        return params

    def _prepare_release(self, payment):
        service = self.client.factory.create('data:CCAuthReversalService')
        service._run = 'true'
        service.authRequestID = payment.transaction_id
        params = {
            'merchantID': self.merchant_id,
            'merchantReferenceCode': payment.id,
            'ccAuthReversalService': service,
            'purchaseTotals': self._prepare_totals(payment)}
        return params

    def _prepare_refund(self, payment, amount=None):
        service = self.client.factory.create('data:CCCreditService')
        service._run = 'true'
        service.captureRequestID = payment.transaction_id
        params = {
            'merchantID': self.merchant_id,
            'merchantReferenceCode': payment.id,
            'ccCreditService': service,
            'purchaseTotals': self._prepare_totals(payment, amount=amount)}
        return params

    def _prepare_card_type(self, card_number):
        card_type, card_name = get_credit_card_issuer(card_number)
        if card_type == 'visa':
            return '001'
        elif card_type == 'mastercard':
            return '002'
        elif card_type == 'amex':
            return '003'
        elif card_type == 'discover':
            return '004'
        elif card_type == 'diners':
            return'005'
        elif card_type == 'jcb':
            return '007'
        elif card_type == 'maestro':
            return '042'

    def _prepare_card_data(self, data):
        card = self.client.factory.create('data:Card')
        card.fullName = data['name']
        card.accountNumber = data['number']
        card.expirationMonth = data['expiration'].month
        card.expirationYear = data['expiration'].year
        card.cvNumber = data['cvv2']
        card.cardType = self._prepare_card_type(data['number'])
        return card

    def _prepare_billing_data(self, payment):
        billing = self.client.factory.create('data:BillTo')
        billing.firstName = payment.billing_first_name
        billing.lastName = payment.billing_last_name
        billing.street1 = payment.billing_address_1
        billing.street2 = payment.billing_address_2
        billing.city = payment.billing_city
        billing.postalCode = payment.billing_postcode
        billing.country = payment.billing_country_code
        billing.state = payment.billing_country_area
        billing.email = payment.billing_email
        billing.ipAddress = payment.customer_ip_address
        return billing

    def _prepare_items(self, payment):
        items = []
        for i, item in enumerate(payment.get_purchased_items()):
            purchased = self.client.factory.create('data:Item')
            purchased._id = i
            purchased.unitPrice = str(item.price)
            purchased.quantity = str(item.quantity)
            purchased.productName = item.name
            purchased.productSKU = item.sku
            items.append(purchased)
        return items

    def _prepare_merchant_defined_data(self, payment):
        try:
            merchant_defined_data = payment.attrs.merchant_defined_data
        except KeyError:
            return
        else:
            data = self.client.factory.create('data:MerchantDefinedData')
            for i, value in merchant_defined_data.items():
                field = self.client.factory.create('data:MDDField')
                field._id = int(i)
                field.value = value
                data.mddField.append(field)
            return data

    def _prepare_totals(self, payment, amount=None):
        totals = self.client.factory.create('data:PurchaseTotals')
        totals.currency = payment.currency
        if amount is None:
            totals.grandTotalAmount = str(payment.total)
            totals.freightAmount = str(payment.delivery)
        else:
            totals.grandTotalAmount = str(amount)
        return totals

    def _serialize_response(self, response):
        if isinstance(response, (Element, Object)):
            response = dict(response)
            for k, v in response.items():
                response[k] = self._serialize_response(v)
        return response

    def process_data(self, payment, request):
        xid = request.POST.get('MD')
        if xid != payment.attrs.xid:
            return redirect(payment.get_failure_url())
        if payment.status in [PaymentStatus.CONFIRMED, PaymentStatus.PREAUTH]:
            return redirect(payment.get_success_url())
        cc_data = request.GET.get('token')
        try:
            cc_data = signing.loads(cc_data)
        except Exception:
            return redirect(payment.get_failure_url())
        else:
            expiration = cc_data['expiration']
            cc_data['expiration'] = datetime.date(
                expiration['year'], expiration['month'], 1)
        params = self._prepare_payer_auth_validation_check(
            payment, cc_data, request.POST.get('PaRes'))
        response = self._make_request(payment, params)
        payment.transaction_id = response.requestID
        try:
            self._set_proper_payment_status_from_reason_code(
                payment, response.reasonCode)
        except PaymentError:
            pass
        if payment.status in [PaymentStatus.CONFIRMED, PaymentStatus.PREAUTH]:
            return redirect(payment.get_success_url())
        else:
            return redirect(payment.get_failure_url())



import re

from django.forms.widgets import MultiWidget
from django.forms.widgets import Select
from django.forms.widgets import TextInput


class CreditCardNumberWidget(TextInput):

    def render(self, name, value, attrs=None, renderer=None):
        if value:
            value = re.sub(r'[\s-]', '', value)
            if len(value) == 16:
                value = ' '.join([value[0:4], value[4:8],
                                  value[8:12], value[12:16]])
            elif len(value) == 15:
                value = ' '.join([value[0:4], value[4:10], value[10:15]])
            elif len(value) == 14:
                value = ' '.join([value[0:4], value[4:10], value[10:14]])
        return super().render(name, value, attrs)


# Credit Card Expiry Fields from:
# http://www.djangosnippets.org/snippets/907/
class CreditCardExpiryWidget(MultiWidget):
    """MultiWidget for representing credit card expiry date."""

    template_name = 'payments/credit_card_expiry_widget.html'

    def decompress(self, value):
        if value:
            return [value.month, value.year]
        else:
            return [None, None]


class SensitiveTextInput(TextInput):
    template_name = 'payments/sensitive_text_input.html'


class SensitiveSelect(Select):
    template_name = 'payments/sensitive_select.html'



import json

import stripe
from django import forms
from django.utils.translation import ugettext as _

from .. import FraudStatus
from .. import PaymentStatus
from .. import RedirectNeeded
from ..forms import CreditCardPaymentFormWithName
from ..forms import PaymentForm as BasePaymentForm
from ..utils import get_month_choices
from ..utils import get_year_choices
from ..widgets import CreditCardExpiryWidget
from ..widgets import SensitiveSelect
from ..widgets import SensitiveTextInput
from .widgets import StripeCheckoutWidget
from .widgets import StripeWidget


class StripeFormMixin:

    charge = None

    def _handle_potentially_fraudulent_charge(self, charge, commit=True):
        fraud_details = charge['fraud_details']
        if fraud_details.get('stripe_report', None) == 'fraudulent':
            self.payment.change_fraud_status(FraudStatus.REJECT, commit=commit)
        else:
            self.payment.change_fraud_status(FraudStatus.ACCEPT, commit=commit)

    def clean(self):
        data = self.cleaned_data

        if not self.errors:
            if not self.payment.transaction_id:
                stripe.api_key = self.provider.secret_key
                try:
                    charge_data = {
                        "capture": False,
                        "amount": int(self.payment.total * 100),
                        "currency": self.payment.currency,
                        "card": data['stripeToken'],
                        "description": '{} {}'.format(
                            self.payment.billing_last_name,
                            self.payment.billing_first_name
                        ),
                    }

                    # Patch charge with billing email if exists
                    if self.payment.billing_email:
                        charge_data.update({
                            "receipt_email": self.payment.billing_email,
                        })

                    self.charge = stripe.Charge.create(**charge_data)

                except stripe.error.CardError as e:
                    # Making sure we retrieve the charge
                    charge_id = e.json_body['error']['charge']
                    self.charge = stripe.Charge.retrieve(charge_id)
                    # Checking if the charge was fraudulent
                    self._handle_potentially_fraudulent_charge(
                        self.charge, commit=False)
                    # The card has been declined
                    self._errors['__all__'] = self.error_class([str(e)])
                    self.payment.change_status(PaymentStatus.ERROR, str(e))
            else:
                msg = _('This payment has already been processed.')
                self._errors['__all__'] = self.error_class([msg])

        return data

    def save(self):
        self.payment.transaction_id = self.charge.id
        self.payment.attrs.charge = json.dumps(self.charge)
        self.payment.change_status(PaymentStatus.PREAUTH)
        if self.provider._capture:
            self.payment.capture()
        # Make sure we store the info of the charge being marked as fraudulent
        self._handle_potentially_fraudulent_charge(self.charge)


class ModalPaymentForm(StripeFormMixin, BasePaymentForm):

    def __init__(self, *args, **kwargs):
        super(StripeFormMixin, self).__init__(hidden_inputs=False, *args, **kwargs)
        widget = StripeCheckoutWidget(provider=self.provider, payment=self.payment)
        self.fields['stripeToken'] = forms.CharField(widget=widget)
        if self.is_bound and not self.data.get('stripeToken'):
            self.payment.change_status(PaymentStatus.REJECTED)
            raise RedirectNeeded(self.payment.get_failure_url())


class PaymentForm(StripeFormMixin, CreditCardPaymentFormWithName):

    stripeToken = forms.CharField(widget=StripeWidget())

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        stripe_attrs = self.fields['stripeToken'].widget.attrs
        stripe_attrs['data-publishable-key'] = self.provider.public_key
        stripe_attrs['data-address-line1'] = self.payment.billing_address_1
        stripe_attrs['data-address-line2'] = self.payment.billing_address_2
        stripe_attrs['data-address-city'] = self.payment.billing_city
        stripe_attrs['data-address-state'] = self.payment.billing_country_area
        stripe_attrs['data-address-zip'] = self.payment.billing_postcode
        stripe_attrs['data-address-country'] = self.payment.billing_country_code
        widget_map = {
            'name': SensitiveTextInput(
                attrs={'autocomplete': 'cc-name', 'required': 'required'}),
            'cvv2': SensitiveTextInput(attrs={'autocomplete': 'cc-csc'}),
            'number': SensitiveTextInput(
                attrs={'autocomplete': 'cc-number', 'required': 'required'}),
            'expiration': CreditCardExpiryWidget(
                widgets=[
                    SensitiveSelect(
                        attrs={'autocomplete': 'cc-exp-month',
                               'required': 'required'},
                        choices=get_month_choices()),
                    SensitiveSelect(
                        attrs={'autocomplete': 'cc-exp-year',
                               'required': 'required'},
                        choices=get_year_choices())])}
        for field_name, widget in widget_map.items():
            self.fields[field_name].widget = widget
            self.fields[field_name].required = False



import hashlib
import hmac
import json
import time
from collections import OrderedDict

import requests
from django.core.exceptions import ImproperlyConfigured
from django.http import HttpResponse
from django.http import HttpResponseForbidden

from .. import PaymentStatus
from ..core import BasicProvider


class CoinbaseProvider(BasicProvider):
    """Payment provider for coinbase.

    This backend implements payments using `Coinbase <https://www.coinbase.com/>`_.

    This backend does not support fraud detection.

    :param key: Api key generated by Coinbase
    :param secret: Api secret generated by Coinbase
    :param endpoint: Coinbase endpoint domain to use. For the production
        environment, use ``'coinbase.com'`` instead
    """

    _method = 'get'
    api_url = 'https://api.%(endpoint)s/v1/buttons'
    checkout_url = 'https://%(endpoint)s/checkouts'

    def __init__(self, key, secret, endpoint='sandbox.coinbase.com', **kwargs):
        self.endpoint = endpoint
        self.key = key
        self.secret = secret
        super().__init__(**kwargs)
        if not self._capture:
            raise ImproperlyConfigured(
                'Coinbase does not support pre-authorization.')

    def get_custom_token(self, payment):
        value = f'coinbase-{payment.token}-{self.key}'
        return hashlib.md5(value.encode('utf-8')).hexdigest()

    def get_checkout_code(self, payment):
        api_url = self.api_url % {'endpoint': self.endpoint}
        button_data = {
            'name': payment.description,
            'price_string': str(payment.total),
            'price_currency_iso': payment.currency,
            'callback_url': self.get_return_url(payment),
            'success_url': payment.get_success_url(),
            'cancel_url': payment.get_failure_url(),
            'custom': self.get_custom_token(payment)}
        # ordered dict for stable JSON output
        data = {'button': OrderedDict(sorted(button_data.items()))}
        nonce = int(time.time() * 1e6)
        message = str(nonce) + api_url + json.dumps(data)
        signature = hmac.new(self.secret.encode(), message.encode(),
                             hashlib.sha256).hexdigest()
        headers = {
            'ACCESS_KEY': self.key,
            'ACCESS_SIGNATURE': signature,
            'ACCESS_NONCE': nonce,
            'Accept': 'application/json'}
        response = requests.post(
            api_url, data=json.dumps(data), headers=headers)

        response.raise_for_status()
        results = response.json()
        return results['button']['code']

    def get_action(self, payment):
        checkout_url = self.checkout_url % {'endpoint': self.endpoint}
        return '{}/{}'.format(checkout_url, self.get_checkout_code(payment))

    def get_hidden_fields(self, payment):
        return {}

    def process_data(self, payment, request):
        try:
            results = json.loads(request.body)
        except (ValueError, TypeError):
            return HttpResponseForbidden('FAILED')

        if results['order']['custom'] != self.get_custom_token(payment):
            return HttpResponseForbidden('FAILED')

        if payment.status == PaymentStatus.WAITING:
            payment.transaction_id = results['order']['transaction']['id']
            payment.change_status(PaymentStatus.CONFIRMED)
            payment.save()
        return HttpResponse('OK')



from decimal import Decimal

from django.core.exceptions import ImproperlyConfigured
from django.http import HttpResponse
from django.http import HttpResponseForbidden

from ..core import BasicProvider
from .forms import ProcessPaymentForm

CENTS = Decimal('0.01')


class DotpayProvider(BasicProvider):
    """Payment provider for dotpay.pl

    This backend implements payments using a popular Polish gateway, `Dotpay.pl
    <http://www.dotpay.pl>`_.

    Due to API limitations there is no support for transferring purchased items.

    This backend does not support fraud detection.

    :param seller_id: Seller ID assigned by Dotpay
    :param pin: PIN assigned by Dotpay
    :param channel: Default payment channel (consult reference guide). Ignored if channel_groups is set.
    :param channel_groups: Payment channels to choose from (consult reference guide). Overrides channel.
    :param lang: UI language
    :param lock: Whether to disable channels other than the default selected above
    :param endpoint: The API endpoint to use. For the production environment, use ``'https://ssl.dotpay.pl/'`` instead
    :param ignore_last_payment_channel: Display default channel or channel groups instead of last used channel.
    :param type: Determines what should be displayed after payment is completed (consult reference guide).
    """
    _method = 'post'

    def __init__(self, seller_id, pin,
                 endpoint='https://ssl.dotpay.pl/test_payment/',
                 channel=0, channel_groups=None, ignore_last_payment_channel=False,
                 lang='pl', lock=False, type=2, **kwargs):
        self.seller_id = seller_id
        self.pin = pin
        self.endpoint = endpoint
        self.channel = channel
        self.channel_groups = channel_groups
        self.ignore_last_payment_channel = ignore_last_payment_channel
        self.lang = lang
        self.lock = lock
        self.type = type
        super().__init__(**kwargs)
        if not self._capture:
            raise ImproperlyConfigured(
                'Dotpay does not support pre-authorization.')

    def get_action(self, payment):
        return self.endpoint

    def get_hidden_fields(self, payment):
        if not payment.description:
            raise ValueError('Payment description is required')

        data = {
            'id': self.seller_id,
            'amount': Decimal(str(payment.total)).quantize(CENTS),
            'control': str(payment.id),
            'currency': payment.currency,
            'description': payment.description,
            'lang': self.lang,
            'ignore_last_payment_channel':
                '1' if self.ignore_last_payment_channel else '0',
            'ch_lock': '1' if self.lock else '0',
            'URL': payment.get_success_url(),
            'URLC': self.get_return_url(payment),
            'type': str(self.type)}
        if self.channel_groups:
            data['channel_groups'] = self.channel_groups
        else:
            data['channel'] = str(self.channel)
        return data

    def process_data(self, payment, request):
        form = ProcessPaymentForm(payment=payment, pin=self.pin,
                                  data=request.POST or None)
        if not form.is_valid():
            return HttpResponseForbidden('FAILED')
        form.save()
        return HttpResponse('OK')



import hashlib
from unittest import TestCase
from unittest.mock import MagicMock
from unittest.mock import Mock

from django.http import HttpResponse
from django.http import HttpResponseForbidden

from . import DotpayProvider
from .. import PaymentStatus
from .forms import COMPLETED
from .forms import REJECTED

VARIANT = 'dotpay'
PIN = '123'
PROCESS_POST = {
    'id': '123',
    'operation_number': 'MI-5',
    'operation_type': 'payment',
    'operation_status': COMPLETED,
    'operation_amount': '212.35',
    'operation_currency': 'PLN',
    'operation_original_amount': '58.07',
    'operation_original_currency': 'USD',
    'operation_datetime': '2018-02-28 12:00:00',
    'control': '1',
    'description': 'Order #1',
    'email': 'user@example.org',
    'p_info': 'John Doe (seller@example.com)',
    'p_email': 'seller@example.com',
    'channel': '1'
}


def get_post_with_sha256(post):
    post = post.copy()
    post['pin'] = '123'
    keys = ['pin',
            'id',
            'operation_number',
            'operation_type',
            'operation_status',
            'operation_amount',
            'operation_currency',
            'operation_original_amount',
            'operation_original_currency',
            'operation_datetime',
            'control',
            'description',
            'email',
            'p_info',
            'p_email',
            'channel']
    key = ''.join([post[key] for key in keys])
    sha256 = hashlib.sha256()
    sha256.update(key.encode('utf-8'))
    key_hash = sha256.hexdigest()
    post['signature'] = key_hash
    return post


class Payment(Mock):
    id = 1
    variant = VARIANT
    currency = 'USD'
    total = 100
    status = PaymentStatus.WAITING

    def get_process_url(self):
        return 'http://example.com'

    def get_failure_url(self):
        return 'http://cancel.com'

    def get_success_url(self):
        return 'http://success.com'

    def change_status(self, status):
        self.status = status


class TestDotpayProvider(TestCase):

    def setUp(self):
        self.payment = Payment()

    def test_get_hidden_fields(self):
        """DotpayProvider.get_hidden_fields() returns a dictionary"""
        provider = DotpayProvider(seller_id='123', pin=PIN)
        self.assertEqual(type(provider.get_hidden_fields(self.payment)), dict)

    def test_process_data_payment_accepted(self):
        """DotpayProvider.process_data() returns a correct HTTP response"""
        request = MagicMock()
        request.POST = get_post_with_sha256(PROCESS_POST)
        params = {
            'seller_id': 123,
            'pin': PIN,
            'endpoint': 'test.endpoint.com',
            'channel': 1,
            'lang': 'en',
            'lock': True}
        provider = DotpayProvider(**params)
        response = provider.process_data(self.payment, request)
        self.assertEqual(type(response), HttpResponse)
        self.assertEqual(self.payment.status, PaymentStatus.CONFIRMED)

    def test_process_data_payment_rejected(self):
        """DotpayProvider.process_data() returns a correct HTTP response"""
        data = dict(PROCESS_POST)
        data.update({'operation_status': REJECTED})
        request = MagicMock()
        request.POST = get_post_with_sha256(data)
        params = {
            'seller_id': 123,
            'pin': PIN,
            'endpoint': 'test.endpoint.com',
            'channel': 1,
            'lang': 'en',
            'lock': True}
        provider = DotpayProvider(**params)
        response = provider.process_data(self.payment, request)
        self.assertEqual(type(response), HttpResponse)
        self.assertEqual(self.payment.status, PaymentStatus.REJECTED)

    def test_incorrect_process_data(self):
        """DotpayProvider.process_data() checks POST signature"""
        request = MagicMock()
        request.POST = PROCESS_POST
        provider = DotpayProvider(seller_id='123', pin=PIN)
        response = provider.process_data(self.payment, request)
        self.assertEqual(type(response), HttpResponseForbidden)

    def test_uses_channel_groups_when_set(self):
        channel_groups = 'K,T'
        params = {
            'seller_id': 123,
            'pin': PIN,
            'endpoint': 'test.endpoint.com',
            'channel': 1,
            'channel_groups': channel_groups,
            'lang': 'en',
            'lock': True}
        provider = DotpayProvider(**params)
        hidden_fields = provider.get_hidden_fields(self.payment)
        self.assertEqual(hidden_fields['channel_groups'], channel_groups)
        self.assertIsNone(hidden_fields.get('channel'))



import re
from typing import Dict
from typing import Tuple
from urllib.parse import urlencode
from urllib.parse import urljoin

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured


PAYMENT_VARIANTS: Dict[str, Tuple[str, Dict]] = {
    'default': ('payments.dummy.DummyProvider', {})}

PAYMENT_HOST = getattr(settings, 'PAYMENT_HOST', None)
if not PAYMENT_HOST:
    if 'django.contrib.sites' not in settings.INSTALLED_APPS:
        raise ImproperlyConfigured('The PAYMENT_HOST setting without '
                                   'the sites app must not be empty.')
    from django.contrib.sites.models import Site

PAYMENT_USES_SSL = getattr(settings, 'PAYMENT_USES_SSL', not settings.DEBUG)


def get_base_url() -> str:
    """
    Returns host url according to project settings. Protocol is chosen by
    checking PAYMENT_USES_SSL variable.
    If PAYMENT_HOST is not specified, gets domain from Sites.
    Otherwise checks if it's callable and returns it's result. If it's not a
    callable treats it as domain.
    """
    protocol = 'https' if PAYMENT_USES_SSL else 'http'
    if not PAYMENT_HOST:
        current_site = Site.objects.get_current()
        domain = current_site.domain
    elif callable(PAYMENT_HOST):
        domain = PAYMENT_HOST()
    else:
        domain = PAYMENT_HOST
    return f'{protocol}://{domain}'


class BasicProvider:
    '''
    This class defines the provider API. It should not be instantiated
    directly. Use factory instead.
    '''
    _method = 'post'

    def get_action(self, payment):
        """The `action` for the HTML form element."""
        return self.get_return_url(payment)

    def __init__(self, capture=True):
        """Create a new provider instance.

        This method should not be called directly; use :func:`provider_factory`
        instead.
        """
        self._capture = capture

    def get_hidden_fields(self, payment):
        '''
        Converts a payment into a dict containing transaction data. Use
        get_form instead to get a form suitable for templates.

        When implementing a new payment provider, overload this method to
        transfer provider-specific data.
        '''
        raise NotImplementedError()

    def get_form(self, payment, data=None):
        '''
        Converts *payment* into a form suitable for Django templates.
        '''
        from .forms import PaymentForm
        return PaymentForm(self.get_hidden_fields(payment),
                           self.get_action(payment), self._method)

    def process_data(self, payment, request):
        """Process callback request from a payment provider.

        This method should handle checking the status of the payment, and
        update the `payment` instance.

        If a client is redirected here after making a payment, then this view
        should redirect them to either :meth:`Payment.get_success_url` or
        :meth:`Payment.get_failure_url`.
        """
        raise NotImplementedError()

    def get_token_from_request(self, payment, request):
        '''
        Return payment token from provider request.
        '''
        raise NotImplementedError()

    def get_return_url(self, payment, extra_data=None):
        """Absolute URL where callbacks are delivered.

        This is the URL where payment providers will forward the user. Many
        payment providers include query params here to facilitate validation
        of the payment.

        Subclasses that redirect the user to the payment provider's page should
        pass this URL as a return/callback URL to providers. Requests to the
        return URL will be passed to the :meth:`~process_data` method.

        Subclasses do not generally need to override this method.
        """
        payment_link = payment.get_process_url()
        url = urljoin(get_base_url(), payment_link)
        if extra_data:
            qs = urlencode(extra_data)
            return url + '?' + qs
        return url

    def capture(self, payment, amount=None):
        raise NotImplementedError()

    def release(self, payment):
        raise NotImplementedError()

    def refund(self, payment, amount=None):
        raise NotImplementedError()


PROVIDER_CACHE = {}


def provider_factory(variant):
    '''
    Return the provider instance based on variant
    '''
    variants = getattr(settings, 'PAYMENT_VARIANTS', PAYMENT_VARIANTS)
    handler, config = variants.get(variant, (None, None))
    if not handler:
        raise ValueError('Payment variant does not exist: %s' %
                         (variant,))
    if variant not in PROVIDER_CACHE:
        module_path, class_name = handler.rsplit('.', 1)
        module = __import__(
            str(module_path), globals(), locals(), [str(class_name)])
        class_ = getattr(module, class_name)
        PROVIDER_CACHE[variant] = class_(**config)
    return PROVIDER_CACHE[variant]


CARD_TYPES = [
    (r'^4[0-9]{12}(?:[0-9]{3,6})?$', 'visa', 'VISA'),
    (r'^(?:5[1-5][0-9]{2}|222[1-9]|22[3-9][0-9]|2[3-6][0-9]{2}|27[01][0-9]|2720)[0-9]{12}$', 'mastercard', 'MasterCard'),
    (r'^6(?:011|5[0-9]{2})[0-9]{12,15}$', 'discover', 'Discover'),
    (r'^3[47][0-9]{13}$', 'amex', 'American Express'),
    (r'^(?:(?:2131|1800|35\d{3})\d{11})$', 'jcb', 'JCB'),
    (r'^(?:3(?:0[0-5]|[68][0-9])[0-9]{11})$', 'diners', 'Diners Club'),
    (r'^(?:5[0678]\d\d|6304|6390|67\d\d)\d{8,15}$', 'maestro', 'Maestro')]


def get_credit_card_issuer(number):
    for regexp, card_type, name in CARD_TYPES:
        if re.match(regexp, number):
            return card_type, name
    return None, None



import json
from decimal import Decimal

import stripe

from .. import PaymentError
from .. import PaymentStatus
from .. import RedirectNeeded
from ..core import BasicProvider
from .forms import ModalPaymentForm
from .forms import PaymentForm


class StripeProvider(BasicProvider):
    """Provider backend using `Stripe <https://stripe.com/>`_.

    This backend does not support fraud detection.

    :param secret_key: Secret key assigned by Stripe.
    :param public_key: Public key assigned by Stripe.
    :param name: A friendly name for your store.
    :param image: Your logo.
    """

    form_class = ModalPaymentForm

    def __init__(self, public_key, secret_key, image='', name='', **kwargs):
        stripe.api_key = secret_key
        self.secret_key = secret_key
        self.public_key = public_key
        self.image = image
        self.name = name
        super().__init__(**kwargs)

    def get_form(self, payment, data=None):
        if payment.status == PaymentStatus.WAITING:
            payment.change_status(PaymentStatus.INPUT)
        form = self.form_class(
            data=data, payment=payment, provider=self)

        if form.is_valid():
            form.save()
            raise RedirectNeeded(payment.get_success_url())
        return form

    def capture(self, payment, amount=None):
        amount = int((amount or payment.total) * 100)
        charge = stripe.Charge.retrieve(payment.transaction_id)
        try:
            charge.capture(amount=amount)
        except stripe.InvalidRequestError:
            payment.change_status(PaymentStatus.REFUNDED)
            raise PaymentError('Payment already refunded')
        payment.attrs.capture = json.dumps(charge)
        return Decimal(amount) / 100

    def release(self, payment):
        charge = stripe.Charge.retrieve(payment.transaction_id)
        charge.refund()
        payment.attrs.release = json.dumps(charge)

    def refund(self, payment, amount=None):
        amount = int((amount or payment.total) * 100)
        charge = stripe.Charge.retrieve(payment.transaction_id)
        charge.refund(amount=amount)
        payment.attrs.refund = json.dumps(charge)
        return Decimal(amount) / 100


class StripeCardProvider(StripeProvider):
    """Provider backend using `Stripe <https://stripe.com/>`_, form-based.

    This backend implements payments using `Stripe <https://stripe.com/>`_ but
    the credit card data is collected by your site.

    Parameters are the same as  :class:`~StripeProvider`.
    """
    form_class = PaymentForm



#!/usr/bin/env python
from setuptools import setup

PACKAGES = [
    "payments",
    "payments.authorizenet",
    "payments.braintree",
    "payments.coinbase",
    "payments.cybersource",
    "payments.dummy",
    "payments.dotpay",
    "payments.paypal",
    "payments.sagepay",
    "payments.sofort",
    "payments.stripe",
    "payments.wallet",
]

REQUIREMENTS = [
    "braintree>=3.14.0",
    "Django>=2.2",
    "cryptography>=1.1.0",
    "PyJWT>=1.3.0,<2.0.0",
    "requests>=1.2.0",
    "stripe>=2.6.0",
    "suds-jurko>=0.6",
    "xmltodict>=0.9.2",
]


setup(
    name="django-payments",
    author="Mirumee Software",
    author_email="hello@mirumee.com",
    description="Universal payment handling for Django",
    long_description=open("README.rst").read(),
    use_scm_version={
        "version_scheme": "post-release",
        "write_to": "payments/version.py",
    },
    setup_requires=["setuptools_scm"],
    url="http://github.com/jazzband/django-payments",
    packages=PACKAGES,
    include_package_data=True,
    classifiers=[
        "Environment :: Web Environment",
        "Framework :: Django :: 2.2",
        "Framework :: Django :: 3.0",
        "Framework :: Django :: 3.1",
        "Framework :: Django :: 3.2",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Framework :: Django",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    install_requires=REQUIREMENTS,
    zip_safe=False,
)



import os

PROJECT_ROOT = os.path.normpath(
    os.path.join(os.path.dirname(__file__), 'payments'))
TEMPLATES = [{
    'BACKEND': 'django.template.backends.django.DjangoTemplates',
    'DIRS': [os.path.join(PROJECT_ROOT, 'templates')]}]

SECRET_KEY = 'NOTREALLY'
PAYMENT_HOST = 'example.com'

INSTALLED_APPS = ['payments', 'django.contrib.sites']



'''
This module is responsible for automatic processing of provider callback
data (asynchronous transaction updates).
'''
from django.db.transaction import atomic
from django.http import Http404
from django.shortcuts import get_object_or_404
from django.urls import path
from django.urls import re_path
from django.views.decorators.csrf import csrf_exempt

from . import get_payment_model
from .core import provider_factory


@csrf_exempt
@atomic
def process_data(request, token, provider=None):
    '''
    Calls process_data of an appropriate provider.

    Raises Http404 if variant does not exist.
    '''
    Payment = get_payment_model()
    payment = get_object_or_404(Payment, token=token)
    if not provider:
        try:
            provider = provider_factory(payment.variant)
        except ValueError:
            raise Http404('No such payment')
    return provider.process_data(payment, request)


@csrf_exempt
@atomic
def static_callback(request, variant):

    try:
        provider = provider_factory(variant)
    except ValueError:
        raise Http404('No such provider')

    token = provider.get_token_from_request(request=request, payment=None)
    if not token:
        raise Http404('Invalid response')
    return process_data(request, token, provider)


urlpatterns = [
    path('process/<uuid:token>/', process_data, name='process_payment'),
    re_path(r'^process/(?P<variant>[a-z-]+)/$', static_callback,
            name='static_process_payment')]



import binascii

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import padding
from cryptography.hazmat.primitives.ciphers import algorithms
from cryptography.hazmat.primitives.ciphers import Cipher
from cryptography.hazmat.primitives.ciphers import modes
from django.core.exceptions import ImproperlyConfigured
from django.shortcuts import redirect

from .. import PaymentStatus
from ..core import BasicProvider


class SagepayProvider(BasicProvider):
    """
    Payment provider for sagepay.com

    This backend implements payments using `SagePay.com <https://www.sagepay.com/>`_
    Form API.

    This backend does not support fraud detection. Purchased items are not currently
    transferred.

    :param vendor: Your vendor code
    :param encryption_key: Encryption key assigned by Sage Pay
    :param endpoint: The API endpoint to use. For the production environment, use ``'https://live.sagepay.com/gateway/service/vspform-register.vsp'`` instead
    """
    _version = '3.00'
    _action = 'https://test.sagepay.com/Simulator/VSPFormGateway.asp'

    def __init__(self, vendor, encryption_key, endpoint=_action, **kwargs):
        self._vendor = vendor
        self._enckey = encryption_key.encode('utf-8')
        self._action = endpoint
        super().__init__(**kwargs)
        if not self._capture:
            raise ImproperlyConfigured(
                'Sagepay does not support pre-authorization.')

    def _get_cipher(self):
        backend = default_backend()
        return Cipher(algorithms.AES(self._enckey), modes.CBC(self._enckey),
                      backend=backend)

    def _get_padding(self):
        return padding.PKCS7(128)

    def aes_enc(self, data):
        data = data.encode('utf-8')
        padder = self._get_padding().padder()
        data = padder.update(data) + padder.finalize()
        encryptor = self._get_cipher().encryptor()
        enc = encryptor.update(data) + encryptor.finalize()
        return b"@" + binascii.hexlify(enc)

    def aes_dec(self, data):
        data = data.lstrip(b'@')
        data = binascii.unhexlify(data)
        decryptor = self._get_cipher().decryptor()
        data = decryptor.update(data) + decryptor.finalize()
        return data.decode('utf-8')

    def get_hidden_fields(self, payment):
        payment.save()
        return_url = self.get_return_url(payment)
        data = {
            'VendorTxCode': payment.pk,
            'Amount': f"{payment.total:.2f}",
            'Currency': payment.currency,
            'Description': f"Payment #{payment.pk}",
            'SuccessURL': return_url,
            'FailureURL': return_url,
            'BillingSurname': payment.billing_last_name,
            'BillingFirstnames': payment.billing_first_name,
            'BillingAddress1': payment.billing_address_1,
            'BillingAddress2': payment.billing_address_2,
            'BillingCity': payment.billing_city,
            'BillingPostCode': payment.billing_postcode,
            'BillingCountry': payment.billing_country_code,
            'DeliverySurname': payment.billing_last_name,
            'DeliveryFirstnames': payment.billing_first_name,
            'DeliveryAddress1': payment.billing_address_1,
            'DeliveryAddress2': payment.billing_address_2,
            'DeliveryCity': payment.billing_city,
            'DeliveryPostCode': payment.billing_postcode,
            'DeliveryCountry': payment.billing_country_code}
        if payment.billing_country_code == 'US':
            data['BillingState'] = payment.billing_country_area
            data['DeliveryState'] = payment.billing_country_area
        udata = "&".join("%s=%s" % kv for kv in data.items())
        crypt = self.aes_enc(udata)
        return {'VPSProtocol': self._version, 'TxType': 'PAYMENT',
                'Vendor': self._vendor, 'Crypt': crypt}

    def process_data(self, payment, request):
        udata = self.aes_dec(request.GET['crypt'])
        data = {}
        for kv in udata.split('&'):
            k, v = kv.split('=')
            data[k] = v
        success_url = payment.get_success_url()
        if payment.status == PaymentStatus.WAITING:
            # If the payment is not in waiting state, we probably have a page reload.
            # We should neither throw 404 nor alter the payment again in such case.
            if data['Status'] == 'OK':
                payment.captured_amount = payment.total
                payment.change_status(PaymentStatus.CONFIRMED)
                return redirect(success_url)
            else:
                # XXX: We should recognize AUTHENTICATED and REGISTERED in the future.
                payment.change_status(PaymentStatus.REJECTED)
                return redirect(payment.get_failure_url())
        return redirect(success_url)
